// +build windows

package main

import (
	"bytes"
	"crypto/rand"
	"encoding/ascii85"
	"errors"
	"fmt"
	"io"
	"log"
	"os"
	"os/exec"
	"syscall"
	"time"
	"unsafe"

	"golang.org/x/crypto/ssh"
	"golang.org/x/sys/windows"
)

var (
	userenv = windows.NewLazySystemDLL("userenv.dll")

	procCreateProfile = userenv.NewProc("CreateProfile")
	procDeleteProfile = userenv.NewProc("DeleteProfileW")
)

// createProfile, creates the profile and home directory of the user identified
// by Security Identifier sid.
func createProfile(sid, username string) (string, error) {
	const S_OK = 0x00000000
	if err := procCreateProfile.Find(); err != nil {
		return "", err
	}
	psid, err := syscall.UTF16PtrFromString(sid)
	if err != nil {
		return "", err
	}
	pusername, err := syscall.UTF16PtrFromString(username)
	if err != nil {
		return "", err
	}
	var pathbuf [260]uint16
	r1, _, e1 := syscall.Syscall6(procCreateProfile.Addr(), 4,
		uintptr(unsafe.Pointer(psid)),        // _In_  LPCWSTR pszUserSid
		uintptr(unsafe.Pointer(pusername)),   // _In_  LPCWSTR pszUserName
		uintptr(unsafe.Pointer(&pathbuf[0])), // _Out_ LPWSTR  pszProfilePath
		uintptr(len(pathbuf)),                // _In_  DWORD   cchProfilePath
		0, // unused
		0, // unused
	)
	if r1 != S_OK {
		if e1 == 0 {
			return "", os.NewSyscallError("CreateProfile", syscall.EINVAL)
		}
		return "", os.NewSyscallError("CreateProfile", e1)
	}
	profilePath := syscall.UTF16ToString(pathbuf[0:])
	return profilePath, nil
}

// deleteProfile, deletes the profile and home directory of the user identified
// by Security Identifier sid.
func deleteProfile(sid string) error {
	if err := procDeleteProfile.Find(); err != nil {
		return err
	}
	psid, err := syscall.UTF16PtrFromString(sid)
	if err != nil {
		return err
	}
	r1, _, e1 := syscall.Syscall(procDeleteProfile.Addr(), 3,
		uintptr(unsafe.Pointer(psid)), // _In_     LPCTSTR lpSidString,
		0, // _In_opt_ LPCTSTR lpProfilePath,
		0, // _In_opt_ LPCTSTR lpComputerName
	)
	if r1 == 0 {
		if e1 == 0 {
			return os.NewSyscallError("DeleteProfile", syscall.EINVAL)
		}
		return os.NewSyscallError("DeleteProfile", e1)
	}
	return nil
}

func isSpecial(c byte) bool {
	return ('!' <= c && c <= '/') || (':' <= c && c <= '@') ||
		('[' <= c && c <= '`') || ('{' <= c && c <= '~')
}

// validPassword, checks if password s meets the Windows complexity
// requirements defined here:
//
//   https://technet.microsoft.com/en-us/library/hh994562(v=ws.11).aspx
//
func validPassword(s string) bool {
	var (
		digits    bool
		special   bool
		alphaLow  bool
		alphaHigh bool
	)
	if len(s) < 8 {
		return false
	}
	for i := 0; i < len(s); i++ {
		switch c := s[i]; {
		case '0' <= c && c <= '9':
			digits = true
		case 'a' <= c && c <= 'z':
			alphaLow = true
		case 'A' <= c && c <= 'Z':
			alphaHigh = true
		case isSpecial(c):
			special = true
		}
	}
	var n int
	if digits {
		n++
	}
	if special {
		n++
	}
	if alphaLow {
		n++
	}
	if alphaHigh {
		n++
	}
	return n >= 3
}

// generatePassword, returns a 14 char ascii85 encoded password.
func generatePassword() (string, error) {
	const Length = 14

	in := make([]byte, ascii85.MaxEncodedLen(Length))
	if _, err := io.ReadFull(rand.Reader, in); err != nil {
		return "", err
	}
	out := make([]byte, ascii85.MaxEncodedLen(len(in)))
	if n := ascii85.Encode(out, in); n < Length {
		return "", errors.New("short password")
	}

	// replace forward slashes as NET USER does not like them

	var char byte // replacement char
	for _, c := range out {
		if c != '/' {
			char = c
			break
		}
	}
	for i, c := range out {
		if c == '/' {
			out[i] = char
		}
	}
	return string(out[:Length]), nil
}

// randomPassword, returns a ascii85 encoded 14 char password
// if the password is longer than 14 chars NET.exe will ask
// for confirmation due to backwards compatibility issues with
// Windows prior to Windows 2000.
func randomPassword() (string, error) {
	limit := 100
	for ; limit >= 0; limit-- {
		s, err := generatePassword()
		if err != nil {
			return "", err
		}
		if validPassword(s) {
			return s, nil
		}
	}
	return "", errors.New("failed to generate valid Windows password")
}

func userExists(name string) bool {
	_, _, t, err := syscall.LookupSID("", name)
	return err == nil && t == syscall.SidTypeUser
}

func createUserProfile(username, password string) error {
	if userExists(username) {
		return fmt.Errorf("user account already exists: %s", username)
	}

	createCmd := exec.Command("NET.exe", "USER", username, password, "/ADD")
	createOut, err := createCmd.CombinedOutput()
	if err != nil {
		return fmt.Errorf("error creating user (%s): %s", err, string(createOut))
	}

	// Add to Administrators group
	groupCmd := exec.Command("NET.exe", "LOCALGROUP", "Administrators", username, "/ADD")
	groupOut, err := groupCmd.CombinedOutput()
	if err != nil {
		return fmt.Errorf("error adding user to Administrator group (%s): %s",
			err, string(groupOut))
	}

	sid, _, _, err := syscall.LookupSID("", username)
	if err != nil {
		return err
	}
	ssid, err := sid.String()
	if err != nil {
		return err
	}
	_, err = createProfile(ssid, username)
	return err
}

func deleteUserProfile(username string) error {
	sid, _, _, err := syscall.LookupSID("", username)
	if err != nil {
		return err
	}
	ssid, err := sid.String()
	if err != nil {
		return err
	}
	if err := deleteProfile(ssid); err != nil {
		return err
	}

	cmd := exec.Command("NET.exe", "USER", username, "/DELETE")
	out, err := cmd.CombinedOutput()
	if err != nil {
		return fmt.Errorf("error deleting user (%s): %s", err, string(out))
	}

	return nil
}

func LoginSSH(username, password string) error {
	config := &ssh.ClientConfig{
		User: username,
		Auth: []ssh.AuthMethod{
			ssh.Password(password),
		},
		HostKeyCallback: ssh.InsecureIgnoreHostKey(),
	}
	client, err := ssh.Dial("tcp", "localhost:22", config)
	if err != nil {
		return fmt.Errorf("Failed to dial: ", err)
	}
	defer client.Close()

	session, err := client.NewSession()
	if err != nil {
		return fmt.Errorf("Failed to create session: ", err)
	}
	defer session.Close()

	var b bytes.Buffer
	session.Stdout = &b
	if err := session.Run("whoami.exe"); err != nil {
		return fmt.Errorf("Failed to run: " + err.Error())
	}
	return nil
}

func randomUsername() string {
	username := fmt.Sprintf("ssh_%d", time.Now().UnixNano())
	if len(username) > 18 {
		username = username[:18]
	}
	return username
}

func realMain() (exitCode int) {
	username := randomUsername()
	fmt.Println("Username:", username)

	password, err := randomPassword()
	if err != nil {
		log.Println("Generating password: %s", err)
		return 1
	}

	if err := createUserProfile(username, password); err != nil {
		log.Println("Creating user (%s): %s", username, err)
		return 1
	}
	fmt.Println("Created user!")

	if err := LoginSSH(username, password); err != nil {
		log.Println("Logging in with user (%s): %s", username, err)
		exitCode = 1
	}
	fmt.Println("Successfully logged in!")

	fmt.Printf("Sleeping for %s before deleting user: %s\n",
		time.Second*5, username)
	time.Sleep(time.Second * 5)

	if err := deleteUserProfile(username); err != nil {
		log.Println("Deleting user (%s): %s", username, err)
		exitCode = 1
	}
	fmt.Println("Successfully deleted user!")

	return exitCode
}

func main() {
	fmt.Println("Creating temp user to log in with...")

	if code := realMain(); code != 0 {
		log.Println("Exiting with error code:", code)
		os.Exit(code)
	}

	fmt.Println("Success!")
}
